{% extends "testable.jinja" %}
{% block imports %}
if 'custom' not in globals():
    from mage_ai.data_preparation.decorators import custom
{% endblock %}

{% block content %}
@custom
def transform_custom(*args, **kwargs):
    """
    Returns:
        Anything (e.g. data frame, dictionary, array, int, str, etc.)
    """
    # Specify your custom logic here
{{ code }}
    return {}
{% endblock %}
